#ifndef _GABRO_H
#define _GABRO_H

#include "DistributionLibrairie.h"
#include "GaBroIdl.h"

using namespace std;

class GaBro : public DistributionLibrary
{
 public:
  GaBro();
  virtual ~GaBro();

  // Common
  virtual void setSourceTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getSourceTopology(); 
  virtual void setDestTopology(PaCO::PacoTopology_t topo); 
  virtual PaCO::PacoTopology_t getDestTopology(); 
  virtual void setNodeRank(long Rank);
  virtual long getNodeRank(); 
  virtual void setCommunicator(void * group) {}

  // Data description API
  void setEltSize(long size);
  void setNodeNbElt(long elt_node_nb);
  void setTotalNbElt(long elt_nb);
  void setNodePos(long node_pos);

  // Client Side
  virtual PieceToSend* computePiecesToSend(unsigned& size);
  virtual void setDataPtr(void* dataPtr);
  virtual void* getClientData(void* pid, int server_node, long & size, long & length, bool & end);
  virtual void clientFree();
  virtual void * getDescr();

  virtual bool setComId(long id);
  virtual bool freeComId(long id);

  // Server Side
  virtual bool insertData(void* rcptBuf, unsigned long element_nb); 
  virtual void* getServerData(long & length);
  virtual void serverFree();
  virtual void setDescr(void * descr);

 private:

  PaCO::PacoTopology_t _srcTopo;
  PaCO::PacoTopology_t _destTopo;

  // Descr
  GaBroLib::GaBroDescr _descr;
  
  // Common
  long _data_size;
  long _data_elt_nb;
  long _my_rank;
  long _server_nb_nodes;
  
  // Client Side
  char * _client_buffer;
  int  * _pos;
  long _elt_node_nb;
  long _node_pos;

  typedef struct {
    long _data_size;
    long _elt_node_nb;
    long _server_nb_nodes;    
    long _data_elt_nb;
    long _node_pos;
  } info_t;

  typedef map<long, info_t*> info_list_map_t; // - 1 = none
  info_list_map_t _info_list_map;
  
  long _cur_id; // - 1 = none
  info_t * _cur_info;
  
  // Server Side
  char * _server_buffer;
  long * _nb_rcp;
  int  _current_rcp;
  long _current_pos;
  long _total_data_nb;
  long _node_data_nb;
  long _current_data_size;
};

class GaBro_fab : virtual public paco_fabrique_distribution
{

 public:
  GaBro_fab() {}
  ~GaBro_fab() {}
  
  DistributionLibrary * create()
    {
      return new GaBro();
    }
};

#endif
