#include "Identity.h"
#include <iostream>

#define SRC_TOPOLOGY 1
#define DST_TOPOLOGY 2
#define ELEMENT_SIZE 4
#define LOCAL_SIZE   8

#define CORRECT_CONFIG 15

#undef DEBUG_INTERNAL
#undef DEBUG_INTERNAL2

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

Identity::Identity()
{
  // Common
  _config = 0;

  // By default 1->1 :)
  _sTopo.total=1;
  _dTopo.total=1;
  _nodeRank=0;
  
  // Client side
  _clientBuffer=NULL;

  // Server side
  _serverDescr=NULL;
  _serverBuffer=NULL;
}

Identity::~Identity() 
{
}

void
Identity::setSourceTopology(PaCO::PacoTopology_t topo)
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- setSourceTopology: " << topo.total << std::endl;
#endif
  _config |= SRC_TOPOLOGY;
  _sTopo = topo;
}

PaCO::PacoTopology_t
Identity::getSourceTopology()
{
  std::cerr << "****** Why is the method " << __FUNCTION__ << "called?\n";
  abort();
  return _sTopo;
}

void
Identity::setDestTopology(PaCO::PacoTopology_t topo)
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- setDestTopology: " << topo.total << std::endl;
#endif
  _config |= DST_TOPOLOGY;
  _dTopo = topo;

}

PaCO::PacoTopology_t
Identity::getDestTopology()
{
  std::cerr << "****** Why is the method " << __FUNCTION__ << "called?\n";
  abort();
  return _dTopo;
}

void
Identity::setNodeRank(long Rank)
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- setNodeRank: " << Rank << std::endl;
#endif
  _nodeRank = Rank;
}

long
Identity::getNodeRank()
{
  return _nodeRank;
}

void
Identity::setEltSize(unsigned long size)
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- setEltSize: " << size << std::endl;
#endif
  _config |= ELEMENT_SIZE;
  _unitsize = size;
}

void
Identity::setLocalNbElt(unsigned long elt_nb)
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- setLocalNbElt: " << elt_nb << std::endl;
#endif
  _config |= LOCAL_SIZE;
  _clientDescr.llen = _llen = elt_nb;
}
  

PieceToSend* 
Identity::computePiecesToSend(unsigned& size_out)
{

  PieceToSend * sched;

#ifdef DEBUG_INTERNAL
  std::cerr << "computePiecesToSend-------------------- in\n";
  std::cerr << __FUNCTION__ << ": config = "<<_config<<endl;
#endif

  if (_config != CORRECT_CONFIG) {
    std::cerr << "Identity: incorrect configuration state "<< _config<<"/"<<CORRECT_CONFIG<<endl;
    abort();
  }

  sched = new PieceToSend[1];
  size_out = 1;

  sched[0].sourceNode = _nodeRank; // my self :)
  sched[0].destNode   = _nodeRank; // the corresponding node
  sched[0].size       = _llen;     // msg size
  sched[0].id         = (void*) -1; /* not used */
 

#ifdef DEBUG_INTERNAL
  std::cerr << "computePiecesToSend-------------------- done\n";
#endif

  return sched;
}

void
Identity::setDataPtr(void* dataPtr)
{
  _clientBuffer = (char *) dataPtr;
#ifdef DEBUG_INTERNAL
  cerr << "-- _clientBuffer set to "<<dataPtr<<endl;
#endif
}

void *
Identity::getClientData(void *pid, int dnode, long & remaining_size_octet, long & returned_length_element, bool & end)  
{
  if (remaining_size_octet < _llen) {
    std::cerr << "Identity: not yet implemented !\n";
    abort();
  }
    
#ifdef DEBUG_INTERNAL
  std::cerr << __FUNCTION__ << ": getClientData returns #elt: "<<_llen<<endl;
#endif

  remaining_size_octet-= _llen;
  returned_length_element= _llen;
  end=true;

  return _clientBuffer;
}

void
Identity::clientFree() // always called at the end of an invocation (reset some variables)
{

#ifdef DEBUG_INTERNAL
  std::cerr << "-- clientFree\n";
#endif
}

void *
Identity::getDescr()
{
#ifdef DEBUG_INTERNAL
  std::cerr << "-- descr:llen: "<<_clientDescr.llen<<std::endl;
#endif
  return &_clientDescr;
}

bool
Identity::insertData(void* rcptBuf, unsigned long element_nb)
{
  // cp from recpBuf to _serverBuffer
#ifdef DEBUG_INTERNAL
  std::cerr << "-- insertData "<<endl;
#endif
  
  if (element_nb != _llen) {
    std::cerr << "Identity: not yet implemented !\n";
    abort();
  }

  unsigned long llen = element_nb*_unitsize;
  memcpy(_serverBuffer, rcptBuf, llen);

  return true;
}

void *
Identity::getServerData(long & length) // in element !
{
  length= _llen;
#ifdef DEBUG_INTERNAL
  std::cerr << "-- getServerData: ptr : "<<(void*)_serverBuffer<<" #element:"<<length<<std::endl;
#endif  
  return (void*) _serverBuffer;
}

void
Identity::setDescr(void * descr)
{
  std::cerr << "-- insertData "<<endl;

  _serverDescr = (IdentityLib::IdentityDescr*) descr;

  // Descr already received

  if ( _serverBuffer == NULL)
    {
      // 1st time -> allocate memory & init
      //this->setEltSize(_descr->usz); BUG: usz can be different in client and server (32bit vs 64bit)
      unsigned long len = _serverDescr->llen*_unitsize;
      this->setLocalNbElt(_serverDescr->llen);
      
#ifdef DEBUG_INTERNAL
      fprintf(stderr, "-- setDescr: stopo: %ld\tdtopo: %ld\n",_sTopo.total, _dTopo.total);
      fprintf(stderr, "   setDescr: local len: %ld", _serverDescr->llen);
#endif  
      _serverBuffer = (char*) malloc(len);
#ifdef DEBUG_INTERNAL
      std::cerr << "-- serverMalloc: "<<(void*)_serverBuffer<<" - "<<_llen<<endl;
#endif
      _serverToReceived = len;
    }
}

void
Identity::serverFree() 
{ 
#ifdef DEBUG_INTERNAL
  std::cerr << "-- serverFree: "<<(void*)_serverBuffer<<endl;
#endif
  free(_serverBuffer);
  _serverBuffer=NULL; // the orb (or the user) need to free the memory!
}

